import numpy as np
import pandas as pd
from datetime import datetime
from scipy import integrate, signal, stats
from tqdm import tqdm
from multiprocessing import cpu_count
try:
    import statsmodels.formula.api as smf
    from pathos.multiprocessing import ProcessingPool as Pool
    import fbm
    import PyVSL
except:
    pass

from . import utils
from .proxy import ProxyRecord

def clean_df(df, mask=None):
    pd.options.mode.chained_assignment = None
    if mask is not None:
        df_cleaned = df.loc[mask]
    else:
        df_cleaned = df

    for col in df.columns:
        df_cleaned.dropna(subset=[col], inplace=True)

    return df_cleaned

class TempPlusNoise:
    ''' A PSM that adds noise to the input climate signal.

    Args:
        pobj (cfr.proxy.ProxyRecord): the proxy record object
        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
    '''
    def __init__(self, pobj=None, climate_required=['tas']):
        self.pobj = pobj
        self.climate_required = climate_required

    def calibrate(self, SNR=10, seasonality=list(range(1, 13)), vn='model.tas',
                  seed=None, noise='white', colored_noise_kws=None):
        ''' Calibrate the PSM.'''
        sigma = np.nanstd(self.pobj.clim[vn].da.values) / SNR
        if noise == 'white':
            rng = np.random.default_rng(seed)
            self.noise = rng.normal(0, sigma, np.size(self.pobj.clim[vn].da.values))
        elif noise == 'colored':
            colored_noise_kws = {} if colored_noise_kws is None else colored_noise_kws
            _colored_noise_kws = {'seed': seed, 'alpha': 1, 't': self.pobj.clim[vn].da.time}
            _colored_noise_kws.update(colored_noise_kws)
            self.noise = utils.colored_noise(**_colored_noise_kws)

        calib_details = {
            'PSMmse': np.mean(self.noise**2),
            'SNR': SNR,
            'seasonality': seasonality,
        }
        self.calib_details = calib_details

    def forward(self, vn='model.tas', no_noise=False):
        ''' Forward the PSM.'''
        if no_noise:
            value = self.pobj.clim[vn].da.values 
        else:
            value=self.pobj.clim[vn].da.values + self.noise

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=self.pobj.clim[vn].da.time,
            value=value,
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
            seasonality=self.calib_details['seasonality'],
        )

        return pp

class Linear:
    ''' A PSM that is based on univariate linear regression.

    Args:
        pobj (cfr.proxy.ProxyRecord): the proxy record object
        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
    '''
    def __init__(self, pobj=None, climate_required=['tas']):
        self.pobj = pobj
        self.climate_required = climate_required

    def calibrate(self, calib_period=None, nobs_lb=25, metric='fitR2adj',
        season_list=[list(range(1, 13))], exog_name=None, **fit_args):
        if exog_name is None:
            exog_name = f'obs.{self.climate_required[0]}'

        exog = self.pobj.clim[exog_name]

        if type(season_list[0]) is not list:
            season_list = [season_list]

        score_list = []
        mdl_list = []
        df_list = []
        sn_list = []
        exog_colname = exog_name.split('.')[-1]

        for sn in season_list:
            exog_ann = exog.annualize(months=sn)
            df_exog = pd.DataFrame({'time': exog_ann.da.time.values, exog_colname: exog_ann.da.values})
            df_proxy = pd.DataFrame({'time': self.pobj.time, 'proxy': self.pobj.value})
            df = df_proxy.dropna().merge(df_exog.dropna(), how='inner', on='time')
            df.set_index('time', drop=True, inplace=True)
            df.sort_index(inplace=True)
            df.astype(float)
            if calib_period is not None:
                mask = (df.index>=calib_period[0]) & (df.index<=calib_period[1])
                df = clean_df(df, mask=mask)

            formula_spell = f'proxy ~ {exog_colname}'
            nobs = len(df)
            if nobs < nobs_lb:
                print(f'The number of overlapped data points is {nobs} < {nobs_lb}. Skipping ...')
            else:
                mdl = smf.ols(formula=formula_spell, data=df).fit(**fit_args)
                fitR2adj =  mdl.rsquared_adj,
                mse = np.mean(mdl.resid**2),
                score = {
                    'fitR2adj': fitR2adj,
                    'mse': mse,
                }
                score_list.append(score[metric])
                mdl_list.append(mdl)
                df_list.append(df)
                sn_list.append(sn)

        if len(score_list) > 0:
            opt_idx_dict = {
                'fitR2adj': np.argmax(score_list),
                'mse': np.argmin(score_list),
            }

            opt_idx = opt_idx_dict[metric]
            opt_mdl = mdl_list[opt_idx]
            opt_sn = sn_list[opt_idx]

            calib_details = {
                'df': df_list[opt_idx],
                'nobs': opt_mdl.nobs,
                'fitR2adj': opt_mdl.rsquared_adj,
                'PSMresid': opt_mdl.resid,
                'PSMmse': np.mean(opt_mdl.resid**2),
                'SNR': np.std(opt_mdl.predict()) / np.std(opt_mdl.resid),
                'seasonality': opt_sn,
            }

            self.calib_details = calib_details
            self.model = opt_mdl
        else:
            self.calib_details = None
            self.model = None

    def forward(self, exog_name=None):
        if exog_name is None:
            exog_name = f'model.{self.climate_required[0]}'

        sn = self.calib_details['seasonality']
        exog = self.pobj.clim[exog_name]
        exog = exog.annualize(months=sn)
        exog_colname = exog_name.split('.')[-1]
        exog_dict = {
            exog_colname: exog.da.values,
        }

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=exog.da.time.values,
            value=np.array(self.model.predict(exog=exog_dict).values),
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
            seasonality=self.calib_details['seasonality'],
        )

        return pp


class Bilinear:
    ''' A PSM that is based on bivariate linear regression.

    Args:
        pobj (cfr.proxy.ProxyRecord): the proxy record object
        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
    '''
    def __init__(self, pobj=None, climate_required = ['tas', 'pr']):
        self.pobj = pobj
        self.climate_required = climate_required

    def calibrate(self, calib_period=None, nobs_lb=25, metric='fitR2adj',
        season_list1=[list(range(1, 13))], season_list2=[list(range(1, 13))],
        exog1_name=None, exog2_name=None, **fit_args):

        if exog1_name is None:
            exog1_name = f'obs.{self.climate_required[0]}'

        if exog2_name is None:
            exog2_name = f'obs.{self.climate_required[1]}'

        exog1 = self.pobj.clim[exog1_name]
        exog2 = self.pobj.clim[exog2_name]

        score_list = []
        mdl_list = []
        df_list = []
        sn_list = []
        exog1_colname = exog1_name.split('.')[-1]
        exog2_colname = exog2_name.split('.')[-1]

        if type(season_list1[0]) is not list:
            season_list1 = [season_list1]

        if type(season_list2[0]) is not list:
            season_list2 = [season_list2]

        for sn1 in season_list1:
            exog1_ann = exog1.annualize(months=sn1)
            df_exog1 = pd.DataFrame({'time': exog1_ann.da.time.values, exog1_colname: exog1_ann.da.values})
            df_proxy = pd.DataFrame({'time': self.pobj.time, 'proxy': self.pobj.value})
            df = df_proxy.dropna().merge(df_exog1.dropna(), how='inner', on='time')
            df_copy = df.copy()
            for sn2 in season_list2:
                exog2_ann = exog2.annualize(months=sn2)
                df_exog2 = pd.DataFrame({'time': exog2_ann.da.time.values, exog2_colname: exog2_ann.da.values})
                df = df.merge(df_exog2.dropna(), how='inner', on='time')
                df.set_index('time', drop=True, inplace=True)
                df.sort_index(inplace=True)
                df.astype(float)
                if calib_period is not None:
                    mask = (df.index>=calib_period[0]) & (df.index<=calib_period[1])
                    df = clean_df(df, mask=mask)

                formula_spell = f'proxy ~ {exog1_colname} + {exog2_colname}'
                nobs = len(df)
                if nobs < nobs_lb:
                    print(f'The number of overlapped data points is {nobs} < {nobs_lb}. Skipping ...')
                else:
                    mdl = smf.ols(formula=formula_spell, data=df).fit(**fit_args)
                    fitR2adj =  mdl.rsquared_adj,
                    mse = np.mean(mdl.resid**2),
                    score = {
                        'fitR2adj': fitR2adj,
                        'mse': mse,
                    }
                    score_list.append(score[metric])
                    mdl_list.append(mdl)
                    df_list.append(df)
                    sn_list.append((sn1, sn2))

                df = df_copy.copy()

        if len(score_list) > 0:
            opt_idx_dict = {
                'fitR2adj': np.argmax(score_list),
                'mse': np.argmin(score_list),
            }

            opt_idx = opt_idx_dict[metric]
            opt_mdl = mdl_list[opt_idx]
            opt_sn = sn_list[opt_idx]

            calib_details = {
                'df': df_list[opt_idx],
                'nobs': opt_mdl.nobs,
                'fitR2adj': opt_mdl.rsquared_adj,
                'PSMresid': opt_mdl.resid,
                'PSMmse': np.mean(opt_mdl.resid**2),
                'SNR': np.std(opt_mdl.predict()) / np.std(opt_mdl.resid),
                'seasonality': opt_sn,
                # 'score_list': score_list,  # for debug purpose
            }

            self.calib_details = calib_details
            self.model = opt_mdl
        else:
            self.calib_details = None
            self.model = None

    def forward(self, exog1_name=None, exog2_name=None):
        if exog1_name is None:
            exog1_name = f'model.{self.climate_required[0]}'
        if exog2_name is None:
            exog2_name = f'model.{self.climate_required[1]}'

        exog1 = self.pobj.clim[exog1_name]
        exog2 = self.pobj.clim[exog2_name]
        sn1, sn2 = self.calib_details['seasonality']
        exog1 = exog1.annualize(months=sn1)
        exog2 = exog2.annualize(months=sn2)

        common_yrs = np.intersect1d(exog1.da.time.values, exog2.da.time.values)
        exog1.da = exog1.da.sel(time=common_yrs)
        exog2.da = exog2.da.sel(time=common_yrs)

        exog1_colname = exog1_name.split('.')[-1]
        exog2_colname = exog2_name.split('.')[-1]
        exog_dict = {
            exog1_colname: exog1.da.values,
            exog2_colname: exog2.da.values,
        }

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=exog1.da.time.values,
            value=np.array(self.model.predict(exog=exog_dict).values),
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
            seasonality=self.calib_details['seasonality'],
        )

        return pp


class Ice_d18O():
    ''' The ice core d18O model adopted from PRYSM (https://github.com/sylvia-dee/PRYSM)

    Args:
        pobj (cfr.proxy.ProxyRecord): the proxy record object
        climate_required (cfr.climate.ClimateField): the required climate field object for running this PSM
    '''
    def __init__(self, pobj=None, tas_name='model.tas', pr_name='model.pr', psl_name='model.psl', d18O_name='model.d18O',
                 climate_required=['tas', 'pr', 'psl', 'd18O']):
        self.pobj = pobj
        self.tas_name = tas_name
        self.pr_name = pr_name
        self.psl_name = psl_name
        self.d18O_name = d18O_name
        self.climate_required = climate_required

    def forward(self, alt_diff=0, nproc=None):
        ''' The ice d18O model

        It takes model simulated montly tas, pr, psl, d18O as input.
        '''
        def ice_sensor(year, d18Op, pr, alt_diff=0.):
            ''' Icecore sensor model

            The ice core sensor model calculates precipitation-weighted del18OP (i.e. isotope ratio is weighted by
            the amount of precipitation that accumulates) and corrects for temperature and altitude bias between model
            and site ([Yurtsever, 1975], 0.3/100m [Vogel et al., 1975]).

            Args:
                year (1d array: time): time axis [year in float]
                d18Op (2d array: location, time): d18O of precipitation [permil]
                pr (2d array: location, time): precipitation rate [kg m-2 s-1]
                alt_diff (2d array: location): actual Altitude-Model Altitude [meters]

            Returns:
                d18Oice (2d array: location, year in int): annualizd d18O of ice [permil]

            References:
                Yurtsever, Y., Worldwide survey of stable isotopes in precipitation., Rep. Isotope Hydrology Section, IAEA, 1975.

            '''
            # Altitude Effect: cooling and precipitation of heavy isotopes.
            # O18 ~0.15 to 0.30 permil per 100m.

            alt_eff = -0.25
            alt_corr = (alt_diff/100.)*alt_eff

            year_ann, d18Op_weighted = utils.annualize_var(year, d18Op, weights=pr)
            d18O_ice = d18Op_weighted + alt_corr

            return d18O_ice


        def diffusivity(rho, T=250, P=0.9, rho_d=822, b=1.3):
            '''
            DOCSTRING: Function 'diffusivity'
            Description: Calculates diffusivity (in m^2/s) as a function of density.

            Inputs:
            P: Ambient Pressure in Atm
            T: Temperature in K
            rho: density profile (kg/m^3)
            rho_d: 822 kg/m^2 [default], density at which ice becomes impermeable to diffusion

            Defaults are available for all but rho, so only one argument need be entered.

            Note values for diffusivity in air:

            D16 = 2.1e-5*(T/273.15)^1.94*1/P
            D18 = D16/1.0285
            D2 = D16/1.0251
            D17 = D16/((D16/D18)^0.518)

            Reference: Johnsen et al. (2000): Diffusion of Stable isotopes in polar firn and ice:
            the isotope effect in firn diffusion

            '''

            # Set Constants

            R = 8.314478                                                # Gas constant
            m = 18.02e-3                                                # molar weight of water (in kg)
            alpha18 = np.exp(11.839/T-28.224e-3)                    # ice-vapor fractionation for oxygen 18
            p = np.exp(9.5504+3.53*np.log(T)-5723.265/T-0.0073*T)     # saturation vapor pressure
            Po = 1.                                                 # reference pressure, atmospheres
            rho_i = 920.  # kg/m^3, density of solid ice

            # Set diffusivity in air (units of m^2/s)

            Da = 2.1e-5*np.power((T/273.15), 1.94)*(Po/P)
            Dai = Da/1.0285

            # Calculate Tortuosity

            invtau = np.zeros(len(rho))
            #  for i in range(len(rho)):
            #      if rho[i] <= rho_i/np.sqrt(b):
            #          # invtau[i]=1.-1.3*np.power((rho[i]/rho_d),2)
            #          invtau[i] = 1.-1.3*np.power((rho[i]/rho_i), 2)
            #      else:
            #          invtau[i] = 0.

            selector =rho <= rho_i/np.sqrt(b)
            invtau[selector] = 1.-1.3*(rho[selector]/rho_i)**2

            D = m*p*invtau*Dai*(1/rho-1/rho_d)/(R*T*alpha18)

            return D


        def densification(Tavg, bdot, rhos, z):  # ,model='hljohnsen'):
            ''' Calculates steady state snow/firn depth density profiles using Herron-Langway type models.



            Args:
                Tavg: 10m temperature in celcius ## CELCIUS!  # fzhu: should be in K now
                bdot: accumulation rate in mwe/yr or (kg/m2/yr)
                rhos: surface density in kg/m3
                z: depth in true_metres

                model can be: {'HLJohnsen' 'HerronLangway' 'LiZwally' 'Helsen' 'NabarroHerring'}
                default is herronlangway. (The other models are tuned for non-stationary modelling (Read Arthern et al.2010 before applying in steady state).

            Returns:
                rho: density (kg/m3) for all z-values.
                zieq: ice equivalent depth for all z-values.
                t: age for all z-values (only taking densification into account.)

                Example usage:
                z=0:300
                [rho,zieq,t]=densitymodel(-31.5,177,340,z,'HerronLangway')
                plot(z,rho)

            References:
                Herron-Langway type models. (Arthern et al. 2010 formulation).
                Aslak Grinsted, University of Copenhagen 2010
                Adapted by Sylvia Dee, Brown University, 2017
                Optimized by Feng Zhu, University of Southern California, 2017

            '''
            rhoi = 920.
            rhoc = 550.
            rhow = 1000.
            rhos = 340.
            R = 8.314

            # Tavg=248.
            # bdot=0.1
            # Herron-Langway with Johnsen et al 2000 corrections.
            # Small corrections to HL model which are not in Arthern et al. 2010

            c0 = 0.85*11*(bdot/rhow)*np.exp(-10160./(R*Tavg))
            c1 = 1.15*575*np.sqrt(bdot/rhow)*np.exp(-21400./(R*Tavg))

            k0 = c0/bdot  # ~g4
            k1 = c1/bdot

            # critical depth at which rho=rhoc
            zc = (np.log(rhoc/(rhoi-rhoc))-np.log(rhos/(rhoi-rhos)))/(k0*rhoi)  # g6

            ix = z <= zc  # find the z's above and below zc
            upix = np.where(ix)  # indices above zc
            dnix = np.where(~ix)  # indices below zc

            q = np.zeros((z.shape))  # pre-allocate some space for q, rho
            rho = np.zeros((z.shape))

            # test to ensure that this will not blow up numerically if you have a very very long core.
            # manually set all super deep layers to solid ice (rhoi=920)
            NUM = k1*rhoi*(z-zc)+np.log(rhoc/(rhoi-rhoc))

            numerical = np.where(NUM <= 100.0)
            blowup = np.where(NUM > 100.0)

            q[dnix] = np.exp(k1*rhoi*(z[dnix]-zc)+np.log(rhoc/(rhoi-rhoc)))  # g7
            q[upix] = np.exp(k0*rhoi*z[upix]+np.log(rhos/(rhoi-rhos)))  # g7

            rho[numerical] = q[numerical]*rhoi/(1+q[numerical])  # [g8] modified by fzhu to fix inconsistency of array size
            rho[blowup] = rhoi

            # only calculate this if you want zieq
            tc = (np.log(rhoi-rhos)-np.log(rhoi-rhoc))/c0  # age at rho=rhoc [g17]
            t = np.zeros((z.shape))  # pre allocate a vector for age as a function of z
            t[upix] = (np.log(rhoi-rhos)-np.log(rhoi-rho[upix]))/c0  # [g16] above zc
            t[dnix] = (np.log(rhoi-rhoc)-np.log(rhoi+0.0001-rho[dnix]))/c1 + tc  # [g16] below zc
            tdiff = np.diff(t)

            # make sure time keeps increasing even after we reach the critical depth.
            if np.any(tdiff == 0.00):
                inflection = np.where(tdiff == 0.0)
                lineardepth_change = t[inflection][0]

                for i in range(len(t)):
                    if t[i] > lineardepth_change:
                        t[i] = t[i-1] + 1e-5

            zieq = t*bdot/rhoi  # [g15]

            return rho, zieq, t


        def ice_archive(d18Oice, pr_ann, tas_ann, psl_ann, nproc=8):
            ''' Accounts for diffusion and compaction in the firn.

            Args:
                d18Oice (1d array: year in int): annualizd d18O of ice [permil]
                pr_ann (1d array: year in int): precipitation rate [kg m-2 s-1]
                tas_ann (1d array: year in int): annualizd atomspheric temerature [K]
                psl_ann (1d array: year in int): annualizd sea level pressure [Pa]
                nproc (int): the number of processes for multiprocessing

            Returns:
                ice_diffused (1d array: year in int): archived ice d18O [permil]

            '''
            # ======================================================================
            # A.0: Initialization
            # ======================================================================
            # accumulation rate [m/yr]
            # note that the unit of pr_ann is [kg m-2 s-1], so need to divide by density [kg m-3] and convert the time
            yr2sec_factor = 3600*24*365.25
            accum = pr_ann/1000*yr2sec_factor

            # depth horizons (accumulation per year corresponding to depth moving down-core)
            bdown = accum[::-1]
            bmean = np.mean(bdown)
            depth = np.sum(bdown)
            depth_horizons = np.cumsum(bdown)
            dz = np.min(depth_horizons)/10.  # step in depth [m]

            Tmean = np.mean(tas_ann)  # unit in [K]
            Pmean = np.mean(psl_ann)*9.8692e-6  # unit in [Atm]

            # contants
            rho_s = 300.  # kg/m^3, surface density
            rho_d = 822.  # kg/m^2, density at which ice becomes impermeable to diffusion
            rho_i = 920.  # kg/m^3, density of solid ice

            # ======================================================================
            # A.1: Compaction Model
            # ======================================================================
            z = np.arange(0, depth, dz) + dz  # linear depth scale

            # set density profile by calling densification function
            rho, zieq, t = densification(Tmean, bmean, rho_s, z)

            rho = rho[:len(z)]  # cutoff the end
            time_d = np.cumsum(dz/bmean*rho/rho_i)
            ts = time_d*yr2sec_factor  # convert time in years to ts in seconds

            # integrate diffusivity along the density gradient to obtain diffusion length
            D = diffusivity(rho, Tmean, Pmean, rho_d, bmean)

            D = D[:-1]
            rho = rho[:-1]
            diffs = np.diff(z)/np.diff(time_d)
            diffs = diffs[:-1]

            # Integration using the trapezoidal method

            # IMPORTANT: once the ice reaches crtiical density (solid ice), there will no longer
            # be any diffusion. There is also numerical instability at that point. Set Sigma=1E-13 for all
            # points below that threshold.

            # Set to 915 to be safe.
            solidice = np.where(rho >= rho_d-5.0)
            diffusion = np.where(rho < rho_d-5.0)

            dt = np.diff(ts)
            sigma_sqrd_dummy = 2*np.power(rho, 2)*dt*D
            sigma_sqrd = integrate.cumtrapz(sigma_sqrd_dummy)
            diffusion_array = diffusion[0]
            diffusion_array = diffusion_array[diffusion_array < len(sigma_sqrd)]  # fzhu: to avoid the boundary index error
            diffusion = np.array(diffusion_array)

            #  rho=rho[0:-1] # modified by fzhu to fix inconsistency of array size
            #  sigma=np.zeros((len(rho)+1)) # modified by fzhu to fix inconsistency of array size
            sigma = np.zeros((len(rho)))
            sigma[diffusion] = np.sqrt(1/np.power(rho[diffusion],2)*sigma_sqrd[diffusion]) # modified by fzhu to fix inconsistency of array size
            #sigma[solidice]=np.nanmax(sigma) #max diffusion length in base of core // set in a better way. max(sigma)
            sigma[solidice] = sigma[diffusion][-1]
            sigma = sigma[:-1]

            # ======================================================================
            # A.2. Diffusion Profile
            # ======================================================================
            # Load water isotope series
            del18 = np.flipud(d18Oice)  # NOTE YOU MIGHT NOT NEED FLIP UD here. Our data goes forward in time.

            # interpolate over depths to get an array of dz values corresponding to isotope values for convolution/diffusion
            iso_interp = np.interp(z, depth_horizons, del18)

            # Return a warning if the kernel length is approaching 1/2 that of the timeseries.
            # This will result in spurious numerical effects.

            zp = np.arange(-100, 100, dz)
            if (len(zp) >= 0.5*len(z)):
                print("Warning: convolution kernel length (zp) is approaching that of half the length of timeseries. Kernel being clipped.")
                bound = 0.20*len(z)*dz
                zp = np.arange(-bound, bound, dz)

            #  print('start for loop ...')
            #  start_time = time.time()

            rm = np.nanmean(iso_interp)
            cdel = iso_interp-rm

            diffused_final = np.zeros(len(iso_interp))
            if nproc == 1:
                for i in tqdm(range(len(sigma))):
                    sig = sigma[i]
                    part1 = 1./(sig*np.sqrt(2.*np.pi))
                    part2 = np.exp(-zp**2/(2*sig**2))
                    G = part1*part2
                    #  diffused = np.convolve(G, cdel, mode='same')*dz  # fzhu: this is way too slow
                    diffused = signal.fftconvolve(cdel, G, mode='same')*dz  # put cdel in the front to keep the same length as before
                    diffused += rm  # remove mean and then put back
                    diffused_final[i] = diffused[i]

            else:
                #  print('Multiprocessing: nproc = {}'.format(nproc))

                def conv(sig, i):
                    part1 = 1./(sig*np.sqrt(2.*np.pi))
                    part2 = np.exp(-zp**2/(2*sig**2))
                    G = part1*part2
                    diffused = signal.fftconvolve(cdel, G, mode='same')*dz
                    diffused += rm  # remove mean and then put back

                    return diffused[i]

                res = Pool(nproc).map(conv, sigma, range(len(sigma)))
                diffused_final[:len(res)] = np.array(res)

            #  print('for loop: {:0.2f} s'.format(time.time()-start_time))

            # take off the first few and last few points used in convolution
            diffused_timeseries = diffused_final[0:-3]

            # Now we need to pack our data back into single year data units based on the depths and year interpolated data
            final_iso = np.interp(depth_horizons, z[0:-3], diffused_timeseries)
            ice_diffused = final_iso

            return ice_diffused

        ################
        # run the model

        # annualize the data
        tas_ann = self.pobj.clim[self.tas_name].annualize()
        pr_ann = self.pobj.clim[self.pr_name].annualize()
        psl_ann = self.pobj.clim[self.psl_name].annualize()

        year_in_float = utils.datetime2year_float(self.pobj.clim[self.pr_name].da.time.values)
        # sensor model
        d18O_ice = ice_sensor(
            year_in_float,
            self.pobj.clim[self.d18O_name].da.values,
            self.pobj.clim[self.pr_name].da.values,
            alt_diff=alt_diff,
        )

        # diffuse model
        if nproc is None: nproc = cpu_count()
        ice_diffused = ice_archive(d18O_ice, pr_ann.da.values, tas_ann.da.values, psl_ann.da.values, nproc=nproc)

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=tas_ann.da.time.values,
            value=ice_diffused[::-1],
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )

        return pp

class Lake_VarveThickness():
    ''' The varve thickness model.

    It takes summer temperature as input (JJA for NH and DJF for SH).
    '''
    def __init__(self, pobj=None, model_tas_name='model.tas', climate_required=['tas']):
        self.pobj = pobj
        self.model_tas_name = model_tas_name
        self.climate_required = climate_required

    def forward(self, H=None, shape=None, mean=None, SNR=None, seed=None, season=None):

        def simpleVarveModel(signal, H=0.75, shape=1.5, mean=1, SNR=.25, seed=0):
            ''' The python version of the simple varve model (in R lang) by Nick McKay
                Adapted by Feng Zhu

            Args:
                signal (array): the signal matrix, each row is a time series
                H (float): Hurst index, should be in (0, 1)
            '''
            # first, gammify the input
            gamSig = gammify(signal, shape=shape, mean=mean, seed=seed)

            # create gammafied autocorrelated fractal brownian motion series
            #     wa = Spectral.WaveletAnalysis()
            #     fBm_ts = wa.fBMsim(N=np.size(signal), H=H)
            fBm_ts = fbm.fgn(np.size(signal), H)
            gamNoise = gammify(fBm_ts, shape=shape, mean=mean, seed=seed)

            # combine the signal with the noise, based on the SNR
            varves = (gamSig*SNR + gamNoise*(1/SNR)) / (SNR + 1/SNR)

            res = {
                'signal': signal,
                'gamSig': gamSig,
                'fBm_ts': fBm_ts,
                'gamNoise': gamNoise,
                'varves': varves,
                'H': H,
                'shape': shape,
                'mean': mean,
                'SNR': SNR,
                'seed': seed,
            }

            return res

        def gammify(X, shape=1.5, mean=1, jitter=False, seed=None):
            ''' Transform each **row** of data matrix X to a gaussian distribution using the inverse Rosenblatt transform
            '''
            X = np.matrix(X)
            n = np.shape(X)[0]  # number of rows
            p = np.shape(X)[1]  # number of columns

            if seed is not None:
                np.random.seed(seed)

            if jitter:
                # add tiny random numbers to aviod ties
                X += np.random.normal(0, np.std(X)/1e6, n*p).reshape(n, p)

            Xn = np.matrix(np.zeros(n*p).reshape(n, p))
            for j in range(n):
                # sort the data in ascending order and retain permutation indices
                R = stats.rankdata(X[j, :])
                # the cumulative distribution function
                CDF = R/p - 1/(2*p)
                # apply the inverse Rosenblatt transformation
                rate = shape/mean
                Xn[j, :] = stats.gamma.ppf(CDF, shape, scale=1/rate)  # Xn is now gamma distributed

            return Xn
        ###########################

        # update the params
        kwargs = {
            'H': 0.75,
            'shape': 1.5,
            'mean': 1,
            'SNR': 0.25,
            'seed': None,
            'season': [6, 7, 8]
        }

        input_args = {
            'H': H,
            'shape': shape,
            'mean': mean,
            'SNR': SNR,
            'seed': seed,
            'season': season,
        }

        for k, v in kwargs.items():
            if input_args[k] is not None:
                kwargs[k] = input_args[k]

        # run the model
        tas_ann = self.pobj.clim[self.model_tas_name].annualize(kwargs['season'])
        varve_res = simpleVarveModel(
            tas_ann.da.values,
            H=kwargs['H'],
            shape=kwargs['shape'],
            mean=kwargs['mean'],
            SNR=kwargs['SNR'],
            seed=kwargs['seed'],
        )

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=tas_ann.da.time.values,
            value=np.array(varve_res['varves'])[0],
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )

        return pp

class Coral_SrCa:
    ''' The coral Sr/Ca model
    '''
    def __init__(self, pobj=None, model_tos_name='model.tos', climate_required='tos'):
        self.pobj = pobj
        self.model_tos_name = model_tos_name
        self.climate_required = climate_required

    def forward(self, b=10.553, a=None, seed=None):
        ''' Sensor model for Coral Sr/Ca = a * tos + b

        Args:
            tos (1-D array): sea surface temperature in [degC]
        '''
        if a is None:
            mu = -0.06
            std = 0.01
            a = stats.norm.rvs(loc=mu, scale=std, random_state=seed)

        SrCa = a*self.pobj.clim[self.model_tos_name].da.values + b

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=self.pobj.clim[self.model_tos_name].da.time.values,
            value=SrCa,
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )

        return pp

class Coral_d18O:
    ''' The PSM is based on the forward model published by [Thompson, 2011]:
       <Thompson, D. M., T. R. Ault, M. N. Evans, J. E. Cole, and J. Emile-Geay (2011),
       Comparison of observed and simulated tropical climate trends using a forward
       model of coral d18O, Geophys.Res.Lett., 38, L14706, doi:10.1029/2011GL048224.>
       Returns a numpy array that is the same size and shape as the input vectors for SST, SSS.
    '''
    def __init__(self, pobj=None, model_tos_name='model.tos', model_d18Osw_name='model.d18Osw',
        climate_required=['tos', 'd18Osw'], species='default'):
        self.pobj = pobj
        self.model_tos_name = model_tos_name
        self.model_d18Osw_name = model_d18Osw_name
        self.species = species
        self.climate_required = climate_required

    def forward(self, b1=0.3007062, b2=0.2619054, b3=0.436509, b4=0.1552032, b5=0.15):
        def pseudocoral(sst, sss=None, d18O=None, species="default", lat=None, lon=None, b1=b1, b2=b2, b3=b3, b4=b4, b5=b5):

            """
            DOCSTRING: Function 'pseudocoral' produces a d18O-coral record given SST, SSS, and global position.
               The function is based on the forward model published by [Thompson, 2011]:
               <Thompson, D. M., T. R. Ault, M. N. Evans, J. E. Cole, and J. Emile-Geay (2011),
               Comparison of observed and simulated tropical climate trends using a forward
               model of coral \u03b418O, Geophys.Res.Lett., 38, L14706, doi:10.1029/2011GL048224.>
               Returns a numpy array that is the same size and shape as the input vectors for SST, SSS.

            Input parameters:
                Latitude    [lat]       [-90, 90]
                Longitude   [lon]       [0, 360]
                SST         [SST]       SST ANOMALY Units = degrees Celsius
                SSS         [SSS]       SSS ANOMALY Units = PSU
                Please note that input files must be read in as 1-D vectors.

            Output:
                Returns a Numpy Array, ['coral'], which is saved in the main program.
                Optional Inputs: Please note that the function will use default params for
                d18O vs. SSS, a, and b unless specified and placed in the appropriate location as shown below.
                delta_18O [permil]: If the user does not have an input field for d18O, you must put -1 in its
                position in the call and use the equation for SSS below.
                **Note: [always use d18O (seawater) if available!]**
                The user may specify values for parameters a and b in the coral d18O forward model,
                which are currently set to the following default values:
                a = coral - SST regression, accounting for species-dependent variations in the SST-18O dependence.
                The code assigns this value universally at -.22, but published values vary by species/genus (i.e. Moses, 2006)
                Specify keyword input 'species=' to set slope:
                species =
                        "Default":      a = -0.22
                        "Porites_sp":   a = -.26178
                        "Porites_lob":  a = -.19646
                        "Porites_lut":  a = -.17391
                        "Porites_aus":  a = -.21
                        "Montast":      a = -.22124
                        "Diploas":      a = -.14992
                b = d18O-SSS slope, as defined by location in: Legrande & Schmidt, (2006)
                        <LeGrande, A. N., and G. A. Schmidt (2006), Global gridded data set of the oxygen isotopic          composition in seawater, Geophys. Res. Lett., 33, L12604, doi:10.1029/2006GL026011.>
                        b1 = 0.31   [Red Sea]
                        b2 = 0.27   [Tropical Pacific]
                        b3 = 0.45   [South Pacific]
                        b4 = 0.16   [Indian Ocean]
                        b5 = 0.15   [Tropical Atlantic Ocean/Caribbean]
            Example of use: Call function 'pseudocoral' to calculate d18O_coral timeseries, lacking d18O sw
            but specifying own numeric parameters for a, b.
            pseudocoral(lat,lon,SST,SSS,-1,species="default",b1,b2,b3,b4,b5)
            """
        #====================================================================



        # Define slope of coral response to SST, a
            #Also, have you considered accounting for species-dependent variations in the a1 value (the SST-18O dependence)?
            #The code assigns this value universally at -.22, but published values vary by species/genus (i.e. Moses, 2006).
            #I've copied and pasted a chunk of code from one of my scripts that does this brute-force... Hope it helps.

            if species == "Porites_sp":
                a=-.26178
            elif species == "Porites_lob":
                a=-.19646
            elif species == "Porites_lut":
                a=-.17391
            elif species == "Porites_aus":
                a=-.21
            elif species== "Montast":
                a=-.22124
            elif species == "Diploas":
                a=-.14992
            elif species == "default":
                a=-0.22
            else:
                a=-0.22

        # Define SSS-d18O(sw) slopes by region. Note that we multiply all slopes by VPDB/VSMOW [0.0020052/0.0020672] ~ 0.97 to correct d18O values.

        # Because d18Osw is reported relative to VSMOW, we have to convert the slope to VPDB so
        # that the units of the forward model output are in VPDB (and thus directly comparable
        # to published d18O records, in units of VPDB). The way this affects the forward model is
        # in the d18Osw-SSS slope.

            V_corr = 0.97002

            slope1 = V_corr*b1      # Red Sea
            slope2 = V_corr*b2      # Tropical Pacific
            slope3 = V_corr*b3      # South Pacific
            slope4 = V_corr*b4      # Indian Ocean
            slope5 = V_corr*b5      # Tropical Atlantic

        # Given Lat/Lon pair, assign correct value to 'b'
            if lon>=32.83 and lon<=43.5 and lat>=12.38 and lat<=28.5:    # Red Sea
                b=slope1
            elif lon<=120.:              # Indian Ocean
                b=slope4
            elif lon>=270. and lat>=10.:    #Tropical Atlantic ~300 E longitude
                b=slope5
            elif lat> -5. and lat<=13.:   # Tropical Pacific
                b=slope2
            elif lat<= -5.:              # South Pacific
                b=slope3
            else:
                b=slope2            # Default = Trop. Pacific.

        #====================================================================

        # Form an array of pseudocoral data
            if np.min(sst) > 200:
                # if in [K], convert to [degC]
                sst -= 273.15

            if d18O is None:
                coral = a*sst+b*sss
            else:
                coral = a*sst + 0.97002*d18O  # SMOW/PDB = 0.97002

        #====================================================================

            return coral

        ####################
        # run the model

        sst = self.pobj.clim[self.model_tos_name]
        d18O = self.pobj.clim[self.model_d18Osw_name]

        value = pseudocoral(
            sst=sst.da.values,
            d18O=d18O.da.values,
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            species=self.species,
            b1=b1,
            b2=b2,
            b3=b3,
            b4=b4,
            b5=b5,
        )

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=sst.da.time.values,
            value=value,
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )

        return pp


class VSLite:
    ''' The VS-Lite tree-ring width model that takes monthly tas, pr as input.
    '''
    def __init__(self, pobj=None,
            obs_tas_name='obs.tas', obs_pr_name='obs.pr',
            model_tas_name='model.tas', model_pr_name='model.pr',
            climate_required=['tas', 'pr'],
        ):
        self.pobj = pobj
        self.obs_tas_name = obs_tas_name
        self.obs_pr_name = obs_pr_name
        self.model_tas_name = model_tas_name
        self.model_pr_name = model_pr_name
        self.climate_required = climate_required

    def calibrate(self, calib_period=[1901, 2000], method='Bayesian', **kwargs):
        proxy_time = self.pobj.time
        proxy_value = self.pobj.value

        obs_tas_time = utils.datetime2year_float(self.pobj.clim[self.obs_tas_name].da.time.values)
        obs_tas_value = self.pobj.clim[self.obs_tas_name].da.values
        if np.min(obs_tas_value) > 200:
            # if in [K], convert to [degC]
            obs_tas_value -= 273.15

        obs_pr_time = utils.datetime2year_float(self.pobj.clim[self.obs_pr_name].da.time.values)
        obs_pr_value = self.pobj.clim[self.obs_pr_name].da.values
        if np.max(obs_pr_value) < 1:
            # if in [kg m-2 s-1], convert to [mm/month]
            obs_pr_value *=  3600*24*30

        proxy_syear = np.min(proxy_time)
        proxy_eyear = np.max(proxy_time)
        obs_syear = np.min(np.floor(obs_tas_time))
        obs_eyear = np.max(np.floor(obs_tas_time))
        calib_syear = np.max([proxy_syear, obs_syear, calib_period[0]])
        calib_eyear = np.min([proxy_eyear, obs_eyear, calib_period[1]])
        if calib_period is not None:
            mask_T = (obs_tas_time>=calib_syear) & (obs_tas_time<calib_eyear)
            mask_P = (obs_pr_time>=calib_syear) & (obs_pr_time<calib_eyear)
            mask_TRW = (proxy_time>=calib_syear) & (proxy_time<calib_eyear)
            T = obs_tas_value[mask_T]
            P = obs_pr_value[mask_P]
            TRW = proxy_value[mask_TRW]
        else:
            T = obs_tas_value
            P = obs_pr_value
            TRW = proxy_value

        if method == 'Bayesian':
            res_dict = PyVSL.est_params(T, P, self.pobj.lat, TRW, **kwargs)
        else:
            raise ValueError('Wrong estimation method for VSLite!')

        self.calib_details = {
            'T1': res_dict['T1'],
            'T2': res_dict['T2'],
            'M1': res_dict['M1'],
            'M2': res_dict['M2'],
            'Rhats': res_dict['Rhats'],
            'convwarning': res_dict['convwarning'],
        }

    def forward(self, **vsl_kwargs):
        tas_time = utils.datetime2year_float(self.pobj.clim[self.model_tas_name].da.time.values)
        syear = np.min(np.floor(tas_time))
        eyear = np.max(np.floor(tas_time))

        if hasattr(self, 'calib_details'):
            kwargs = {}
            for k in ['T1', 'T2', 'M1', 'M2']:
                kwargs[k] = self.calib_details[k]

            vsl_kwargs.update(kwargs)

        model_tas_value = self.pobj.clim[self.model_tas_name].da.values
        model_pr_value = self.pobj.clim[self.model_pr_name].da.values

        if np.min(model_tas_value) > 200:
            # if in [K], convert to [degC]
            model_tas_value -= 273.15

        if np.max(model_pr_value) < 1:
            # if in [kg m-2 s-1], convert to [mm/month]
            model_pr_value *=  3600*24*30

        vsl_res = PyVSL.VSL(
            syear, eyear, self.pobj.lat,
            model_tas_value,
            model_pr_value, **vsl_kwargs
        )

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=np.arange(syear, eyear+1),
            value=vsl_res['trw'],
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )

        return pp


class BayTEX86:
    ''' Baysian PSM for TEX86
    
    References:
        Tierney, J. E. & Tingley, M. P. A Bayesian, spatially-varying calibration model for the TEX86 proxy. Geochimica et Cosmochimica Acta 127, 83-106 (2014).
    '''
    def __init__(self, pobj=None, climate_required=['sst']):
        self.pobj = pobj
        self.climate_required = climate_required

    def forward(self, seed=2333, type='SST', mode='analog', tolerance=1):
        vn = f'model.{self.climate_required[0]}'
        res, sigma = pb.TEX_forward(self.pobj.lat, self.pobj.lon, self.pobj.clim[vn].da.values, seed=seed, type=type, mode=mode, tolerance=tolerance)
        pct = np.percentile(res, q=[5, 50, 95], axis=1)

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=np.array(self.pobj.clim[vn].da.time.values),
            value=pct[1],  # median
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )
        pp.R = sigma**2
        return pp

class BayUK37:
    ''' Baysian PSM for UK37
    
    References:
        Tierney, J. E. & Tingley, M. P. BAYSPLINE: A New Calibration for the Alkenone Paleothermometer. Paleoceanography and Paleoclimatology 33, 281-301 (2018).
    '''
    def __init__(self, pobj=None, climate_required=['sst']):
        self.pobj = pobj
        self.climate_required = climate_required

    def forward(self, seed=2333):
        vn = f'model.{self.climate_required[0]}'
        res, sigma = pb.UK_forward(self.pobj.clim[vn].da.values, seed=seed)
        pct = np.percentile(res, q=[5, 50, 95], axis=1)

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=np.array(self.pobj.clim[vn].da.time.values),
            value=pct[1],  # median
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )
        pp.R = sigma**2
        return pp


class BayD18O:
    ''' Baysian PSM for foram d18O
    
    References:
    '''
    def __init__(self, pobj=None, climate_required=['sst', 'd18Osw']):
        self.pobj = pobj
        self.climate_required = climate_required

    def forward(self, seed=2333, species='all'):
        vn1 = f'model.{self.climate_required[0]}'
        vn2 = f'model.{self.climate_required[1]}'
        res, sigma = pb.d18Oc_forward(sst=self.pobj.clim[vn1].da.values, d18Osw=self.pobj.clim[vn2].da.values, seed=seed, species=species)
        pct = np.percentile(res, q=[5, 50, 95], axis=1)

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=np.array(self.pobj.clim[vn1].da.time.values),
            value=pct[1],  # median
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )
        pp.R = sigma**2
        return pp

class BayMgCa:
    ''' Baysian PSM for Mg/Ca
    
    References:
    '''
    def __init__(self, pobj=None, climate_required=['sst', 'sss']):
        self.pobj = pobj
        self.climate_required = climate_required

    def forward(self, seed=2333, species='all', clean=1, pH=1, omega=1, sw=2, H=1, age=15):
        vn1 = f'model.{self.climate_required[0]}'
        vn2 = f'model.{self.climate_required[0]}'
        res, sigma = pb.MgCa_forward(
            age=age, sst=self.pobj.clim[vn1].da.values, salinity=self.pobj.clim[vn2].da.values,
            clean=clean, pH=pH, omega=omega, species=species, sw=sw, H=H, seed=seed,
        )
        pct = np.percentile(res, q=[5, 50, 95], axis=1)

        pp = ProxyRecord(
            pid=self.pobj.pid,
            time=np.array(self.pobj.clim[vn1].da.time.values),
            value=np.exp(pct[1]),  # median; the raw output is ln(Mg/Ca), so taking exp to recover Mg/Ca
            lat=self.pobj.lat,
            lon=self.pobj.lon,
            elev=self.pobj.elev,
            ptype=self.pobj.ptype,
            value_name=self.pobj.value_name,
            value_unit=self.pobj.value_unit,
            time_name=self.pobj.time_name,
            time_unit=self.pobj.time_unit,
        )
        pp.R = sigma**2
        return pp


import os
from datetime import datetime
import cftime
import xarray as xr
import pandas as pd
import numpy as np
import copy
import plotly.express as px
from tqdm import tqdm
from . import visual
from . import utils
from .ts import EnsTS
import eofs


class ClimateField:
    ''' The class for the gridded climate field data.
    '''
    def __init__(self, da=None):
        ''' Initialize a ClimateField object with a `xarray.DataArray`.

        Args:
            da (xarray.DataArray): the gridded data array.
        '''
        self.da = da

    def __getitem__(self, key):
        ''' This makes the object subscriptable. '''
        if type(key) is int:
            # one index
            da = self.da[key]
        elif type(key) is str:
            # one timestamp
            # key = slice(key, key, None)
            # da = self.da.sel({'time': key})
            # if len(da.values) == 0:
            #     mask = (self.da['time']>=float(key.start)) & (self.da['time']<=float(key.stop))
            #     da = self.da.sel({'time': mask})
            try:
                mask = (utils.datetime2year_float(self.da['time'].values)>=float(key)) & (utils.datetime2year_float(self.da['time'].values)<float(key)+1)
            except:
                mask = (self.da['time'].values>=float(key)) & (self.da['time'].values<float(key)+1)

            da = self.da.sel({'time': mask})

        elif type(key) is list:
            # multiple discrete items
            if type(key[0]) is int:
                # index
                da = self.da[key]
            elif type(key[0]) is str:
                # timestamp
                da_list = []
                for k in key:
                    da_list.append(self[k].da)
                da = xr.concat(da_list, dim='time')
                
        elif type(key) is slice:
            if key.start is not None:
                dtype = type(key.start)
            else:
                key.start = self.da['time'][0]

            if key.stop is not None:
                dtype = type(key.stop)
            else:
                key.stop = self.da['time'][-1]

            if dtype is int:
                # index
                da = self.da[key]
            elif dtype is str:
                # timestamp
                try:
                    mask = (utils.datetime2year_float(self.da['time'].values)>=float(key.start)) & (utils.datetime2year_float(self.da['time'].values)<float(key.stop))
                except:
                    mask = (self.da['time'].values>=float(key.start)) & (self.da['time'].values<float(key.stop))

                da = self.da.sel({'time': mask})
                if key.step is not None:
                    da = da[::int(key.step)]

            else:
                raise TypeError('Wrong type for key!')
        else:
            raise TypeError('Wrong type for key!')

        fd = ClimateField(da)

        return fd

    def __len__(self):
        return len(self.time)

    def wrap_lon(self, mode='360'):
        ''' Convert the longitude values

        Args:
            mode (str): if '360', convert the longitude values from the range (-180, 180) to (0, 360);
                if '180', convert the longitude values from the range (0, 360) to (-180, 180);
        '''
        if mode == '360':
            da = self.da.assign_coords({'lon': np.mod(self.da['lon'], 360)})
        elif mode == '180':
            da = self.da.assign_coords({'lon': ((self.da['lon']+180) % 360)-180})
        else:
            raise ValueError('Wrong mode. Should be either "360" or "180".')

        da = da.sortby(da['lon'])
        fd = ClimateField(da)
        return fd

    def fetch(self, name=None, **load_nc_kws):
        ''' Fetch a gridded climate field from cloud

        Args:
            name (str): a predefined name, or an URL starting with "http", or a local file path.
                If not set, the method will return hints of available predefined names.
            load_nc_kws (dict): the dictionary of keyword arguments for loading a netCDF file.
        '''
        url_dict = utils.climfd_url_dict

        if name is None:
            utils.p_warning(f'>>> Choose one from the supported entries:')
            for k in url_dict.keys():
                utils.p_warning(f'- {k}')
            return None

        if name in url_dict:
            url = url_dict[name]
        else:
            url = name

        if url[:4] == 'http':
            # cloud
            os.makedirs('./data', exist_ok=True)
            fpath = f'./data/{os.path.basename(url)}'
            if os.path.exists(fpath):
                utils.p_hint(f'>>> The target file seems existed at: {fpath} . Loading from it instead of downloading ...')
            else:
                utils.download(url, fpath)
                utils.p_success(f'>>> Downloaded file saved at: {fpath}')
        else:
            # local
            fpath = url

        fd = self.load_nc(fpath, **load_nc_kws)

        return fd

    def from_np(self, time, lat, lon, value):
        ''' Load data from a `numpy.ndarray`.

        Args:
            time (array-like): the array of the time axis.
            lat (array-like): the array of the lat axis.
            lon (array-like): the array of the lon axis.
            value (array-like): the array of the values.
        '''
        lat_da = xr.DataArray(lat, dims=['lat'], coords={'lat': lat})
        lon_da = xr.DataArray(lon, dims=['lon'], coords={'lon': lon})
        time_da = xr.DataArray(time, dims=['time'], coords={'time': time})
        da = xr.DataArray(
            value, dims=['time', 'lat', 'lon'],
            coords={'time': time_da, 'lat': lat_da, 'lon': lon_da},
        )
        fd = ClimateField(da)
        return fd

    def get_anom(self, ref_period=[1951, 1980]):
        ''' Get the anomaly against a reference time period.

        Args:
            ref_period (tuple or list): the reference time period in the form or (start_yr, end_yr)
        '''
        try:
            var_ref = self.da.loc[str(ref_period[0]):str(ref_period[-1])]
        except:
            utils.p_warning(f'>>> The time axis does not overlap with the reference period {ref_period}; use its own time period as reference [{np.min(self.da.time):.2f}, {np.max(self.da.time):.2f}].')
            var_ref = self.da

        clim = var_ref.groupby('time.month').mean('time')
        da = self.da.groupby('time.month') - clim
        fd = ClimateField(da)

        return fd

    def center(self, ref_period=[1951, 1980]):
        ''' Center the climate field against a reference time period.

        Args:
            ref_period (tuple or list): the reference time period in the form or (start_yr, end_yr)
            time_name (str): name of the time dimention
        '''
        try:
            var_ref = self.da.loc[str(ref_period[0]):str(ref_period[-1])]
        except:
            utils.p_warning(f'>>> The time axis does not overlap with the reference period {ref_period}; use its own time period as reference [{np.min(self.da.time):.2f}, {np.max(self.da.time):.2f}].')
            var_ref = self.da

        clim = var_ref.mean('time')
        da = self.da - clim
        fd = ClimateField(da)

        return fd

    def load_nc(self, path, vn=None, time_name='time', lat_name='lat', lon_name='lon', load=True, return_ds=False, use_cftime=True, **kwargs):
        ''' Load the climate field from a netCDF file.

        Args:
            path (str): the path where to load data from.
            vn (str): the variable name to load.
            time_name (str): the name for the time axis. Defaults to 'time'.
            lat_name (str): the name for the lat axis. Defaults to 'lat'.
            lon_name (str): the name for the lon axis. Defaults to 'lon'.
            load (bool): if True, the netCDF file will be loaded into the memory; if False, will take the advantage of lazy loading. Defaults to `False`.
            return_ds (bool): if True, will return a `xarray.Dataset` object instead of a `ClimateField` object. Defaults to `False`.
            use_cftime (bool): if True, use the cftime convention. Defaults to `True`.
        '''
        ds = xr.open_dataset(path, use_cftime=use_cftime, **kwargs)
        if return_ds:
            return ds
        else:
            if len(ds.keys()) == 1:
                vn = list(ds.keys())[0]
            else:
                if vn is None:
                    raise ValueError('Variable name should be specified with `vn`.')

            da = ds[vn]
            if load: da.load()
            fd = ClimateField(da)
            if time_name != 'time':
                fd.da = fd.da.rename({time_name: 'time'})
            if lat_name != 'lat':
                fd.da = fd.da.rename({lat_name: 'lat'})
            if lon_name != 'lon':
                fd.da = fd.da.rename({lon_name: 'lon'})

            if np.min(fd.da.lon) < 0:
                fd = fd.wrap_lon()

        
        return fd

    def to_nc(self, path, verbose=True, compress_params=None):
        ''' Convert the climate field to a netCDF file.

        Args:
            path (str): the path where to save
            verbose (bool, optional): print verbose information. Defaults to False.
            compress_params (dict): the paramters for compression when storing the reconstruction results to netCDF files.
        '''
        _comp_params = {'zlib': True}

        encoding_dict = {}
        if compress_params is not None:
            _comp_params.update(compress_params)

        encoding_dict[self.da.name] = _comp_params

        try:
            dirpath = os.path.dirname(path)
            os.makedirs(dirpath, exist_ok=True)
        except:
            pass

        if os.path.exists(path):
            os.remove(path)

        self.da.to_netcdf(path, encoding=encoding_dict)
        if verbose: utils.p_success(f'>>> ClimateField.da["{self.da.name}"] saved to: {path}')

    def copy(self):
        ''' Make a deepcopy of the object. '''
        return copy.deepcopy(self)

    def rename(self, new_vn):
        ''' Rename the variable name of the climate field.

        Args:
            new_vn (str): the new variable name.
        '''
        da = self.da.rename(new_vn)
        fd = ClimateField(da)
        return fd

    def __add__(self, ref):
        ''' Add a reference.
        '''
        if isinstance(ref, ClimateField):
            da = self.da + ref.da
        elif isinstance(ref, float):
            da = self.da + ref
        elif isinstance(ref, int):
            da = self.da + ref
        else:
            raise ValueError('`ref` should be a `ClimateField` object or a float like value.')

        fd = ClimateField(da)
        return fd

    def __sub__(self, ref):
        ''' Substract a reference.
        '''
        if isinstance(ref, ClimateField):
            da = self.da - ref.da
        elif isinstance(ref, float):
            da = self.da - ref
        elif isinstance(ref, int):
            da = self.da - ref
        else:
            raise ValueError('`ref` should be a `ClimateField` object or a float like value.')

        fd = ClimateField(da)
        return fd

    def __mul__(self, ref):
        ''' Multiply a scaler.
        '''
        if isinstance(ref, ClimateField):
            da = self.da * ref.da
        elif isinstance(ref, float):
            da = self.da * ref
        elif isinstance(ref, int):
            da = self.da * ref
        else:
            raise ValueError('`ref` should be a `ClimateField` object or a float like value.')

        fd = ClimateField(da)
        return fd

    def __truediv__(self, ref):
        ''' Divide a scaler.
        '''
        if isinstance(ref, ClimateField):
            da = self.da / ref.da
        elif isinstance(ref, float):
            da = self.da / ref
        elif isinstance(ref, int):
            da = self.da / ref
        else:
            raise ValueError('`ref` should be a `ClimateField` object or a float like value.')

        fd = ClimateField(da)
        return fd
    
    def __floordiv__(self, ref):
        ''' Floor-divide a scaler.
        '''
        if isinstance(ref, ClimateField):
            da = self.da // ref.da
        elif isinstance(ref, float):
            da = self.da // ref
        elif isinstance(ref, int):
            da = self.da // ref
        else:
            raise ValueError('`ref` should be a `ClimateField` object or a float like value.')

        fd = ClimateField(da)
        return fd

    def compare(self, ref, timespan=None, stat='corr', interp_target='ref', interp=True):
        ''' Compare against a reference field.

        Args:
            ref (cfr.climate.ClimateField): the reference to compare against, assuming the first dimension to be time
            timespan (tuple or list): the timespan over which to compare two ClimateField objects.
            interp_target (str, optional): the direction to interpolate the fields:
            
                * 'ref': interpolate from `self` to `ref`
                * 'self': interpolate from `ref` to `self`

            stat (str): the statistics to calculate. Supported quantaties:

                * 'corr': correlation coefficient
                * 'R2': coefficient of determination
                * 'CE': coefficient of efficiency
        '''
        if interp:
            if interp_target == 'ref':
                fd_rg = self.regrid(ref.da.lat, ref.da.lon)
                ref_rg = ref.copy()
            elif interp_target == 'self':
                fd_rg = self.copy()
                ref_rg = ref.regrid(self.da.lat, self.da.lon)
        else:
            fd_rg = self.copy()
            ref_rg = ref.copy()

        if timespan is not None:
            fd_rg = fd_rg[str(timespan[0]):str(timespan[-1])]
            ref_rg = ref_rg[str(timespan[0]):str(timespan[-1])]

        if len(fd_rg.da.lat.shape) == 1:
            fd_rg.da = xr.DataArray(
                fd_rg.da.values,
                coords={'time': ref_rg.da.time, 'lat': fd_rg.da.lat, 'lon': fd_rg.da.lon}
            )

        if stat == 'corr':
            stat_da = xr.corr(fd_rg.da, ref_rg.da, dim='time')
            stat_da = stat_da.expand_dims({'time': [1]})
            stat_da.name = stat
            stat_fd = ClimateField(stat_da)
            stat_fd.plot_kwargs = {
                'cmap': 'RdBu_r',
                'extend': 'neither',
                'levels': np.linspace(-1, 1, 21),
                'cbar_labels': np.linspace(-1, 1, 11),
                'cbar_title': r'$r$',
                'cbar_title_y': 1,
                'title': 'Correlation',
            }
        elif stat == 'R2':
            ss_res = ((fd_rg.da - ref_rg.da) ** 2).sum(dim='time')
            ss_tot = ((ref_rg.da - ref_rg.da.mean(dim='time')) ** 2).sum(dim='time')
            stat_da = 1 - ss_res/ss_tot
            stat_da = stat_da.expand_dims({'time': [1]})
            stat_da.name = stat
            stat_fd = ClimateField(stat_da)
            stat_fd.plot_kwargs = {
                'cmap': 'Reds',
                'extend': 'min',
                'levels': np.linspace(0, 1, 21),
                'cbar_labels': np.linspace(0, 1, 11),
                'cbar_title': r'$R^2$',
                'cbar_title_y': 1,
                'title': 'Coefficient of Determination',
            }
        elif stat == 'CE':
            ce = utils.coefficient_efficiency(ref_rg.da.values, fd_rg.da.values)
            stat_da = xr.DataArray(
                ce[np.newaxis],
                name=stat,
                coords={
                    'time': [1],
                    'lat': fd_rg.da.lat,
                    'lon': fd_rg.da.lon,
                })
            stat_fd = ClimateField(stat_da)
            stat_fd.plot_kwargs = {
                'cmap': 'RdBu_r',
                'extend': 'min',
                'levels': np.linspace(-1, 1, 21),
                'cbar_labels': np.linspace(-1, 1, 11),
                'cbar_title': r'$CE$',
                'cbar_title_y': 1,
                'title': 'Coefficient of Efficiency',
            }
        else:
            raise ValueError('Wrong `stat`; should be one of `corr`, `R2`, and `CE`.' )

        return stat_fd
            
    def get_eof(self, n=1, time_period=None, verbose=False, flip=False):
        ''' Get the EOF analysis result of the ClimateField object

        Args:
            n (int): perform EOF analysis and return the first `n` modes.
            time_period (tuple or list): the timespan over which to perfom the EOF analysis.
            verbose (bool, optional): print verbose information. Defaults to False.
            flip (bool, optional): flip the sign of the field values. Defaults to False.
        '''
        if time_period is None:
            da = self.da
        else:
            da = self.da.loc[f'{time_period[0]}':f'{time_period[1]}']

        if flip:
            da = -da

        coslat = np.cos(np.deg2rad(self.da['lat'].values))
        wgts = np.sqrt(coslat)[..., np.newaxis]
        solver = eofs.standard.Eof(da.values, weights=wgts)
        modes = solver.eofsAsCorrelation(neofs=n)
        pcs = solver.pcs(npcs=n, pcscaling=1)
        fracs = solver.varianceFraction(neigs=n)

        self.eof_res = {
            'modes': modes,
            'pcs': pcs,
            'fracs': fracs,
            'time_period': time_period,
        }

        if verbose: utils.p_success(f'ClimateField.eof_res created with {n} mode(s).')

    def plot_eof(self, n=1, eof_title=None, pc_title=None):
        ''' Plot the EOF analysis result

        Args:
            n (int): plot the `n`-th mode.
            eof_title (str): the subplot title for the mode field.
            pc_title (str): the subplot title for the PC time series.
        '''
        eof = self.eof_res['modes']
        pc = self.eof_res['pcs']
        fracs = self.eof_res['fracs']
        time_period = self.eof_res['time_period']
        if eof_title is None: eof_title = f'EOF{n} ({fracs[n-1]*100:.2f}%)'
        if pc_title is None: pc_title = f'PC{n} ({fracs[n-1]*100:.2f}%)'

        if time_period is None:
            time = self.time
        else:
            time_mask = (self.time>=time_period[0]) & (self.time<=time_period[1])
            time = self.time[time_mask]

        fig, ax = visual.plot_eof(
            eof[n-1], pc[:, n-1], self.lat, self.lon, time,
            eof_title, pc_title)

        return fig, ax

    def plot(self, **kwargs):
        ''' Plot a climate field at a time point.

        See also:
            cfr.visual.plot_field_map : Visualize a field on a map.
        '''
        cbar_title = visual.make_lb(
            self.da.name,
            self.da.attrs['unit'] if 'unit' in self.da.attrs else None)

        cmap_dict = {
            'tas': 'RdBu_r',
            'tos': 'RdBu_r',
            'sst': 'RdBu_r',
            'pr': 'BrBG',
        }
        cmap = cmap_dict[self.da.name] if self.da.name in cmap_dict else 'viridis'

        _kwargs = {
            'cbar_title': cbar_title,
            'cmap': cmap,
        }

        if hasattr(self, 'plot_kwargs'):
            _kwargs.update(self.plot_kwargs)

        if 'title' not in kwargs and 'title' not in _kwargs:
            if len(self.da.dims) == 3:
                t_value = self.da.time.values[0]
            elif len(self.da.dims) == 2:
                try:
                    t_value = self.da.time.values
                except:
                    t_value = ''

            try:
                date_str = '-'.join(str(t_value).split('-')[:2])
            except:
                date_str = str(t_value)

            kwargs['title'] = f'{self.da.name}, {date_str}' 
            
        # if len(set(np.diff(self.da.lon))) > 1:
        #     # the case when longitudes cross 0 degree
        #     fd_plot = self.wrap_lon(mode='180')
        #     kwargs['central_longitude'] = 0
        # else:
            # fd_plot = self

        fd_plot = self

        _kwargs.update(kwargs)
            
        if len(fd_plot.da.dims) == 3:
            vals = fd_plot.da.values[0]
        elif len(fd_plot.da.dims) == 2:
            vals = fd_plot.da.values

        fig, ax = visual.plot_field_map(vals, fd_plot.da.lat, fd_plot.da.lon, **_kwargs)

        return fig, ax

    def plotly_grid(self, site_lats=None, site_lons=None, **kwargs):
        ''' Plot the grid on an interactive map utilizing Plotly

        Args:
            site_lats (list): a list of the latitudes of the sites to plot
            site_lons (list): a list of the longitudes of the sites to plot
        '''
        nlat, nlon = np.size(self.da.lat), np.size(self.da.lon)
        df = pd.DataFrame()
        n = 0
        for i in range(nlat):
            for j in range(nlon):
                lat = self.da.lat[i]
                lon = self.da.lon[j]
                if not np.isnan(self.da[-1, i, j]):
                    avail = 'Data'
                else:
                    avail = 'NaN'
                df.loc[n, 'lat'] = lat
                df.loc[n, 'lon'] = lon
                df.loc[n, 'Type'] = avail
                n += 1

        if site_lats is not None:
            if type(site_lats) is not list:
                site_lats = [site_lats]

        if site_lons is not None:
            if type(site_lons) is not list:
                site_lons = [site_lons]

        for i, site_lat in enumerate(site_lats):
            for j, site_lon in enumerate(site_lons):
                df.loc[n, 'lat'] = site_lat
                df.loc[n, 'lon'] = site_lon
                df.loc[n, 'Type'] = 'Site'
                n += 1

        fig = px.scatter_geo(
            df, lat='lat', lon='lon',
            color='Type',
            projection='natural earth',
            **kwargs,
        )

        return fig

    def annualize(self, months=list(range(1, 13))):
        ''' Annualize/seasonalize the climate field based on a list of months.

        Args:
            months (list): the months based on which for annualization; e.g., [6, 7, 8] means JJA annualization
        '''
        if 'annualized' in self.da.attrs and self.da.attrs['annualized']==1:
            da = self.da
        else:
            da = utils.annualize(da=self.da, months=months)
            yrs = np.floor(utils.datetime2year_float(da.time.values))
            da = da.assign_coords({'time': [int(y) for y in yrs]})
            da.attrs['annualized'] = 1
        fd = ClimateField(da)
        return fd

    def regrid(self, lats, lons, periodic_lon=False):
        ''' Regrid the climate field.'''

        if len(self.da.lat.values.shape) == 1:
            lat_da = xr.DataArray(lats, dims=['lat'], coords={'lat': lats})
            lon_da = xr.DataArray(lons, dims=['lon'], coords={'lon': lons})
            da = self.da.interp(coords={'lat': lat_da, 'lon': lon_da})

        elif len(self.da.lon.values.shape) == 2:
            rgd_data, lats, lons = utils.regrid_field_curv_rect(
                self.da.values, self.da.lat.values, (self.da.lon.values+180)%360-180,
                lats=lats, lons=lons,
            )
            da = xr.DataArray(
                rgd_data, dims=['time', 'lat', 'lon'],
                coords={
                    'time': self.da.time,
                    'lat': lats,
                    'lon': lons,
                },
                name=self.da.name,
            )

        if periodic_lon:
            da[..., -1] = da[..., 0]

        fd = ClimateField(da)
        return fd

    def crop(self, lat_min=-90, lat_max=90, lon_min=0, lon_max=360):
        ''' Crop the climate field based on the range of latitude and longitude.

        Note that in cases when the crop range is crossing the 0 degree of longitude, `lon_min` should be less than 0.

        Args:
            lat_min (float): the lower bound of latitude to crop.
            lat_max (float): the upper bound of latitude to crop.
            lon_min (float): the lower bound of longitude to crop.
            lon_max (float): the upper bound of longitude to crop.
        '''

        mask_lat = (self.da.lat >= lat_min) & (self.da.lat <= lat_max)
        if lon_min >= 0:
            mask_lon = (self.da.lon >= lon_min) & (self.da.lon <= lon_max)
        else:
            # the case when longitudes in mode [-180, 180]
            lon_min = np.mod(lon_min, 360)
            lon_max = np.mod(lon_max, 360)
            mask_lon = (self.da.lon >= lon_min) | (self.da.lon <= lon_max)

        da = self.da.sel({
                'lat': self.da.lat[mask_lat],
                'lon': self.da.lon[mask_lon],
            })
        fd = ClimateField(da)

        return  fd

    def geo_mean(self, lat_min=-90, lat_max=90, lon_min=0, lon_max=360):
        ''' Calculate the geographical mean value of the climate field.

        Args:
            lat_min (float): the lower bound of latitude for the calculation.
            lat_max (float): the upper bound of latitude for the calculation.
            lon_min (float): the lower bound of longitude for the calculation.
            lon_max (float): the upper bound of longitude for the calculation.
        '''
        fdc = self.crop(lat_min=lat_min, lat_max=lat_max, lon_min=lon_min, lon_max=lon_max)
        wgts = np.cos(np.deg2rad(fdc.da['lat']))
        m = fdc.da.weighted(wgts).mean(('lon', 'lat'))
        ts = EnsTS(time=m['time'].values, value=m.values)
        return ts

    def index(self, name):
        ''' Calculate the predefined indices.

        Args:
            name (str): the predefined index name; supports the below:

                * 'nino3.4'
                * 'nino1+2'
                * 'nino3'
                * 'nino4'
                * 'tpi'
                * 'wp'
                * 'dmi'
                * 'iobw'
        '''
        if name == 'gm': 
            return self.geo_mean()
        elif name == 'nhm': 
            return self.geo_mean(lat_min=0)
        elif name == 'shm':
            return self.geo_mean(lat_max=0)
        elif name == 'nino3.4':
            return self.geo_mean(lat_min=-5, lat_max=5, lon_min=np.mod(-170, 360), lon_max=np.mod(-120, 360))
        elif name == 'nino1+2':
            return self.geo_mean(lat_min=-10, lat_max=10, lon_min=np.mod(-90, 360), lon_max=np.mod(-80, 360))
        elif name == 'nino3':
            return self.geo_mean(lat_min=-5, lat_max=5, lon_min=np.mod(-150, 360), lon_max=np.mod(-90, 360))
        elif name == 'nino4':
            return self.geo_mean(lat_min=-5, lat_max=5, lon_min=np.mod(160, 360), lon_max=np.mod(-150, 360))
        elif name == 'wpi':
            # Western Pacific Index
            return self.geo_mean(lat_min=-10, lat_max=10, lon_min=np.mod(120, 360), lon_max=np.mod(150, 360))
        elif name == 'tpi':
            # Tri-Pole Index
            v1 = self.geo_mean(lat_min=25, lat_max=45, lon_min=np.mod(140, 360), lon_max=np.mod(-145, 360))
            v2 = self.geo_mean(lat_min=-10, lat_max=10, lon_min=np.mod(170, 360), lon_max=np.mod(-90, 360))
            v3 = self.geo_mean(lat_min=-50, lat_max=-15, lon_min=np.mod(150, 360), lon_max=np.mod(-160, 360))
            return v2 - (v1 + v3)/2
        elif name == 'dmi':
            # Indian Ocean Dipole Mode
            dmiw = self.geo_mean(lat_min=-10, lat_max=10, lon_min=50, lon_max=70)
            dmie = self.geo_mean(lat_min=-10, lat_max=0, lon_min=90, lon_max=110)
            return dmiw - dmie
        elif name == 'iobw':
            # Indian Ocean Basin Wide
            return self.geo_mean(lat_min=-20, lat_max=20, lon_min=40 ,lon_max=100)
        else:
            raise ValueError('Wrong index name.')


import xarray as xr
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import cftime
import colorama as ca
import statsmodels.api as sm
import pyresample
import requests
from tqdm import tqdm


def p_header(text):
    # return cprint(text, 'cyan', attrs=['bold'])  # lib: termcolor
    print(ca.Fore.CYAN + ca.Style.BRIGHT + text)
    print(ca.Style.RESET_ALL, end='')

def p_hint(text):
    # return cprint(text, 'grey', attrs=['bold'])  # lib: termcolor
    print(ca.Fore.LIGHTBLACK_EX + ca.Style.BRIGHT + text)
    print(ca.Style.RESET_ALL, end='')

def p_success(text):
    # return cprint(text, 'green', attrs=['bold'])  # lib: termcolor
    print(ca.Fore.GREEN + ca.Style.BRIGHT + text)
    print(ca.Style.RESET_ALL, end='')

def p_fail(text):
    # return cprint(text, 'red', attrs=['bold'])  # lib: termcolor
    print(ca.Fore.RED + ca.Style.BRIGHT + text)
    print(ca.Style.RESET_ALL, end='')

def p_warning(text):
    # return cprint(text, 'yellow', attrs=['bold'])  # lib: termcolor
    print(ca.Fore.YELLOW + ca.Style.BRIGHT + text)
    print(ca.Style.RESET_ALL, end='')

def make_bin_vector(ts, bin_width=10):
    bin_vector = np.arange(
        bin_width*(np.min(ts)//bin_width),
        bin_width*(np.max(ts)//bin_width+2),
        step=bin_width,
    )

    return bin_vector
    
def smooth_ts(ts, ys, bin_vector=None, bin_width=10):
    if bin_vector is None:
        bin_vector = np.arange(
            bin_width*(np.min(ts)//bin_width),
            bin_width*(np.max(ts)//bin_width+2),
            step=bin_width)

    ts_bin = (bin_vector[1:] + bin_vector[:-1])/2
    n_bin = np.size(ts_bin)
    ys_bin = np.zeros(n_bin)

    for i in range(n_bin):
        t_start = bin_vector[i]
        t_end = bin_vector[i+1]
        t_range = (ts >= t_start) & (ts < t_end)
        if np.sum(t_range*1) == 1:
            ys_bin[i] = ys[t_range]
        else:
            ys_bin[i] = np.average(ys[t_range], axis=0)

    return ts_bin, ys_bin, bin_vector

def bin_ts(ts, ys, bin_vector=None, bin_width=10, resolution=1, smoothed=False):
    if not smoothed:
        ts_smooth, ys_smooth, bin_vector = smooth_ts(ts, ys, bin_vector=bin_vector, bin_width=bin_width)
    else:
        ts_smooth, ys_smooth = ts, ys

    bin_vector_finer = np.arange(np.min(bin_vector), np.max(bin_vector)+1, step=resolution)
    bin_value = np.zeros(bin_vector_finer.size)

    n_bin = np.size(ts_smooth)
    for i in range(n_bin):
        t_start = bin_vector[i]
        t_end = bin_vector[i+1]
        t_range = (bin_vector_finer >= t_start) & (bin_vector_finer <= t_end)
        bin_value[t_range] = ys_smooth[i]

    return bin_vector_finer, bin_value

def bootstrap(samples, n_bootstraps=1000, stat_func=np.nanmean):

    n_samples = np.shape(samples)[0]
    stats = np.zeros(n_bootstraps)
    for i in range(n_bootstraps):
        rand_ind = np.random.randint(n_samples, size=n_samples)
        stats[i] = stat_func(samples[rand_ind])

    return stats

def clean_ts(ts, ys):
    ''' Delete the NaNs in the time series and sort it with time axis ascending

    Parameters
    ----------
    ts : array
        The time axis of the time series, NaNs allowed
    ys : array
        A time series, NaNs allowed

    Returns
    -------
    ts : array
        The time axis of the time series without NaNs
    ys : array
        The time series without NaNs
    '''
    ys = np.array(ys, dtype=float)
    ts = np.array(ts, dtype=float)
    assert ys.size == ts.size, 'The size of time axis and data value should be equal!'

    ys_tmp = np.copy(ys)
    ys = ys[~np.isnan(ys_tmp)]
    ts = ts[~np.isnan(ys_tmp)]
    ts_tmp = np.copy(ts)
    ys = ys[~np.isnan(ts_tmp)]
    ts = ts[~np.isnan(ts_tmp)]

    # sort the time series so that the time axis will be ascending
    sort_ind = np.argsort(ts)
    ys = ys[sort_ind]
    ts = ts[sort_ind]

    # handle duplicated time points
    t_count = {}
    value_at_t = {}
    for i, t in enumerate(ts):
        if t not in t_count:
            t_count[t] = 1
            value_at_t[t] = ys[i]
        else:
            t_count[t] += 1
            value_at_t[t] += ys[i]

    ys = []
    ts = []
    for t, v in value_at_t.items():
        ys.append(v / t_count[t])
        ts.append(t)

    ts = np.array(ts)
    ys = np.array(ys)

    return ts, ys


def overlap_ts(ts_proxy, ys_proxy, ts_obs, ys_obs):
    ys_proxy = np.array(ys_proxy, dtype=float)
    ts_proxy = np.array(ts_proxy, dtype=float)
    ys_obs = np.array(ys_obs, dtype=float)
    ts_obs = np.array(ts_obs, dtype=float)

    overlap_proxy = (ts_proxy >= np.min(ts_obs)) & (ts_proxy <= np.max(ts_obs))
    overlap_obs = (ts_obs >= np.min(ts_proxy)) & (ts_obs <= np.max(ts_proxy))

    ys_proxy_overlap, ts_proxy_overlap = ys_proxy[overlap_proxy], ts_proxy[overlap_proxy]
    ys_obs_overlap, ts_obs_overlap = ys_obs[overlap_obs], ts_obs[overlap_obs]

    time_overlap = np.intersect1d(ts_proxy_overlap, ts_obs_overlap)
    ind_proxy = list(i for i, t in enumerate(ts_proxy_overlap) if t in time_overlap)
    ind_obs = list(i for i, t in enumerate(ts_obs_overlap) if t in time_overlap)
    ys_proxy_overlap = ys_proxy_overlap[ind_proxy]
    ys_obs_overlap = ys_obs_overlap[ind_obs]

    return time_overlap, ys_proxy_overlap, ys_obs_overlap

def ols_ts(ts_proxy, ys_proxy, ts_obs, ys_obs):
    ys_proxy = np.array(ys_proxy, dtype=float)
    ts_proxy = np.array(ts_proxy, dtype=float)
    ys_obs = np.array(ys_obs, dtype=float)
    ts_obs = np.array(ts_obs, dtype=float)

    time_overlap, ys_proxy_overlap, ys_obs_overlap = overlap_ts(ts_proxy, ys_proxy, ts_obs, ys_obs)

    # calculate the linear regression
    X = sm.add_constant(ys_proxy_overlap)
    ols_model = sm.OLS(ys_obs_overlap, X, missing='drop')

    return ols_model


def gcd(lat1, lon1, lat2, lon2):
	'''
	Calculate the great circle distance between two points
	on the earth (specified in decimal degrees)

	Parameters:
	-----------
	lat1: float
		Latitude of first location (degrees)
	lon1: float
		Longitude of first location (degrees)
	lat2: float
		Latitude of second location (degrees)
	lon2: float
		Longitude of second location (degrees)
		
	Returns:
	--------
	km: float
		Great circle distance between the two locations (in kilometers)
	'''
	# convert decimal degrees to radians
	lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

	# haversine formula
	dlon = lon2 - lon1
	dlat = lat2 - lat1
	a = np.sin(dlat/2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon/2)**2
	c = 2 * np.arcsin(np.sqrt(a))

	# 6367 km is the radius of the Earth
	km = 6367 * c
	return km

def annualize(time=None, value=None, da=None, months=None):
    # convert negative months to support expressions like [-12, 1, 2]
    months = list(range(1, 13)) if months is None else np.abs(months)

    if da is None:
        dates = year_float2datetime(time)
        da_wk = xr.DataArray(
            value, dims=['time'], coords={'time': dates}
        )
    else:
        da_wk = da

    # sda = da_wk.sel(time=da_wk.time.dt.month.isin(months))
    sda = da_wk.sel(time=da_wk['time.month'].isin(months))
    anchor = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEC']
    idx = months[-1]-1
    sda_ann = sda.resample(time=f'A-{anchor[idx]}').mean()
    time_ann = np.floor(datetime2year_float(sda_ann.time.values))
    value_ann = sda_ann.values

    if da is None:
        return time_ann, value_ann
    else:
        return sda_ann

def annualize_var(year_float, var, resolution='month', weights=None):
    ''' Annualize a variable array
    Args:
        var (ndarray): the target variable array with 1st dim to be year
        year_float (1-D array): the time axis of the variable array
        weights (ndarray): the weights that shares the same shape of the target variable array
    Returns:
        var_ann (ndarray): the annualized variable array
        year_ann (1-D array): the time axis of the annualized variable array
    '''
    var = np.array(var)
    year_float = np.array(year_float)

    ndims = len(np.shape(var))
    dims = ['time']
    for i in range(ndims-1):
        dims.append(f'dim{i+1}')

    time = year_float2datetime(year_float, resolution=resolution)

    if weights is not None:
        weights_da = xr.DataArray(weights, dims=dims, coords={'time': time})

        coeff = np.ndarray(np.shape(weights))
        for i, gp in enumerate(list(weights_da.groupby('time.year'))):
            year, value = gp
            k = np.shape(value)[0]
            coeff[k*i:k*(i+1)] = value / np.sum(value, axis=0)

        del weights, weights_da  # save the memory

        var = np.multiply(coeff, var)
        var_da = xr.DataArray(var, dims=dims, coords={'time': time})
        var_ann = var_da.groupby('time.year').sum('time')

    else:
        var_da = xr.DataArray(var, dims=dims, coords={'time': time})
        var_ann = var_da.groupby('time.year').mean('time')

    var_ann = var_ann.values

    year_ann = np.sort(list(set([t.year for t in time])))
    return year_ann, var_ann

def ymd2year_float(year, month, day, resolution='month'):
    ''' Convert a set of (year, month, day) to an array of floats in unit of year
    '''
    year_float = []
    for y, m, d in zip(year, month, day):
        if resolution == 'month': d = 1
        date = datetime(year=y, month=m, day=d)
        fst_day = datetime(year=y, month=1, day=1)
        lst_day = datetime(year=y+1, month=1, day=1)
        year_part = date - fst_day
        year_length = lst_day - fst_day
        year_float.append(y + year_part/year_length)

    year_float = np.array(year_float)
    return year_float

def datetime2year_float(date, resolution='month'):
    ''' Convert a list of dates to floats in year
    '''
    if isinstance(date[0], np.datetime64):
        date = pd.to_datetime(date)

    year = [d.year for d in date]
    month = [d.month for d in date]
    day = [d.day for d in date]
    if resolution == 'month': day = np.ones(np.size(day))

    year_float = ymd2year_float(year, month, day, resolution)

    return year_float

def year_float2datetime(year_float, resolution='month'):
    ''' Convert an array of floats in unit of year to a datetime time; accuracy: one day
    '''
    # if np.min(year_float) < 0:
    #     raise ValueError('Cannot handel negative years. Please truncate first.')

    year = np.array([int(y) for y in year_float], dtype=int)
    month = np.zeros(np.size(year), dtype=int)
    day = np.zeros(np.size(year), dtype=int)

    for i, y in enumerate(year):
        fst_day = datetime(year=y, month=1, day=1)
        lst_day = datetime(year=y+1, month=1, day=1)
        year_length = lst_day - fst_day

        year_part = (year_float[i] - y)*year_length + timedelta(minutes=1)  # to fix the numerical error
        date = year_part + fst_day
        month[i] = date.month
        day[i] = date.day

    if resolution == 'day':
        time = [cftime.datetime(y, m, d, 0, 0, 0, 0, 0, 0, calendar='standard') for y, m, d in zip(year, month, day)]
    elif resolution == 'month':
        time = [cftime.datetime(y, m, 1, 0, 0, 0, 0, 0, 0, calendar='standard') for y, m in zip(year, month)]

    return time

def year_float2dates(year_float):
    ''' Convert an array of floats in unit of year to a datetime time; accuracy: one day
    '''
    # if np.min(year_float) < 0:
    #     raise ValueError('Cannot handel negative years. Please truncate first.')

    year = np.array([int(y) for y in year_float], dtype=int)
    dates = []

    for i, y in enumerate(year):
        fst_day = datetime(year=y, month=1, day=1)
        lst_day = datetime(year=y+1, month=1, day=1)
        year_length = lst_day - fst_day

        year_part = (year_float[i] - y)*year_length + timedelta(minutes=1)  # to fix the numerical error
        date = year_part + fst_day
        dates.append(date)


    return dates

def regrid_field(field, lat, lon, lat_new, lon_new):
    import spharm
    nlat_old, nlon_old = np.size(lat), np.size(lon)
    nlat_new, nlon_new = np.size(lat_new), np.size(lon_new)
    spec_old = spharm.Spharmt(nlon_old, nlat_old, gridtype='regular', legfunc='computed')
    spec_new = spharm.Spharmt(nlon_new, nlat_new, gridtype='regular', legfunc='computed')

    field_new = []
    for field_old in field:
        regridded_field =  spharm.regrid(spec_old, spec_new, field_old, ntrunc=None, smooth=None)
        field_new.append(regridded_field)

    field_new = np.array(field_new)
    return field_new

def regrid_field_curv_rect(field, lat_curv, lon_curv,
        lats=None, lons=None,
        dlat=1, dlon=1, lon_range=None, lat_range=None,
        roi=None, roi_factor=1e6):
    ''' Regrid a curvilinear grid to a linear rectilinear grid

    Args:
        field (np.ndarray): assuming shape in (..., nlat, nlon)

    Note: lon_curv should be in range (-180, 180)!
    '''
    if lons is None:
        if lon_range is not None:
            lons = np.arange(lon_range[0], lon_range[-1]+dlon, dlon)
        else:
            lons = np.arange(np.min(lon_curv), np.max(lon_curv)+dlon, dlon)

    dlon = np.abs(np.median(np.diff(lons)))

    if lats is None:
        if lat_range is not None:
            lats = np.arange(lat_range[0], lat_range[-1]+dlat, dlat)
        else:
            lats = np.arange(np.min(lat_curv), np.max(lat_curv)+dlat, dlat)

    # convert the longitude from (0, 360) to (-180, 180)
    if np.min(lon_curv) >= 0:
        lon_curv = (lon_curv+180) % 360 -180

    lon_rect, lat_rect = np.meshgrid(lons, lats)
    lon_rect = pyresample.utils.wrap_longitudes(lon_rect)

    if roi is None:
        roi = roi_factor * dlon

    old_grid = pyresample.geometry.SwathDefinition(lons=lon_curv, lats=lat_curv)
    new_grid = pyresample.geometry.SwathDefinition(lons=lon_rect, lats=lat_rect)

    fd_shape = field.shape
    field_resh = field.reshape(-1, fd_shape[-2], fd_shape[-1]).transpose(1, 2, 0)
    res = pyresample.kd_tree.resample_nearest(
        old_grid, field_resh, new_grid,
        radius_of_influence=roi,
        fill_value=np.nan,
    )
    rgd_data = res.transpose(2, 0, 1).reshape(*fd_shape[:-2], len(lats), len(lons))

    return rgd_data, lats, lons

def coefficient_efficiency(ref, test, valid=None):
    """ Compute the coefficient of efficiency for a test time series, with respect to a reference time series.

    Args:
        ref (numpy.ndarray):   reference array, of same size as test
        test (numpy.ndarray):  test array
        valid (float): fraction of valid data required to calculate the statistic

    Note:
        Assumes that the first dimension in test and ref arrays is time!!!

    Returns:
        CE (float): CE statistic calculated following Nash & Sutcliffe (1970)
    """

    # check array dimensions
    dims_test = test.shape
    dims_ref  = ref.shape
    # print('dims_test: ', dims_test, ' dims_ref: ', dims_ref)

    if len(dims_ref) == 3:   # 3D: time + 2D spatial
        dims = dims_ref[1:3]
    elif len(dims_ref) == 2: # 2D: time + 1D spatial
        dims = dims_ref[1:2]
    elif len(dims_ref) == 1: # 0D: time series
        dims = 1
    else:
        print('In coefficient_efficiency(): Problem with input array dimension! Exiting...')
        SystemExit(1)

    CE = np.zeros(dims)

    # error
    error = test - ref

    # CE
    numer = np.sum(np.power(error,2),axis=0)
    denom = np.sum(np.power(ref-np.nanmean(ref,axis=0),2),axis=0)
    CE    = 1. - np.divide(numer,denom)

    if valid:
        nbok  = np.sum(np.isfinite(ref),axis=0)
        nball = float(dims_ref[0])
        ratio = np.divide(nbok,nball)
        indok  = np.where(ratio >= valid)
        indbad = np.where(ratio < valid)
        dim_indbad = len(indbad)
        testlist = [indbad[k].size for k in range(dim_indbad)]
        if not all(v == 0 for v in testlist):
            if isinstance(dims,(tuple,list)):
                CE[indbad] = np.nan
            else:
                CE = np.nan

    return CE

def geo_mean(da, lat_min=-90, lat_max=90, lon_min=0, lon_max=360, lat_name='lat', lon_name='lon'):
    ''' Calculate the geographical mean value of the climate field.

    Args:
        lat_min (float): the lower bound of latitude for the calculation.
        lat_max (float): the upper bound of latitude for the calculation.
        lon_min (float): the lower bound of longitude for the calculation.
        lon_max (float): the upper bound of longitude for the calculation.
    '''
    # calculation
    mask_lat = (da[lat_name] >= lat_min) & (da[lat_name] <= lat_max)
    mask_lon = (da[lon_name] >= lon_min) & (da[lon_name] <= lon_max)

    dac = da.sel(
        {
            lat_name: da[lat_name][mask_lat],
            lon_name: da[lon_name][mask_lon],
        }
    )

    wgts = np.cos(np.deg2rad(dac[lat_name]))
    m = dac.weighted(wgts).mean((lon_name, lat_name))
    return m

def colored_noise(alpha, t, f0=None, m=None, seed=None):
    ''' Generate a colored noise timeseries

    Parameters
    ----------
    alpha : float
        exponent of the 1/f^alpha noise
    t : float
        time vector of the generated noise
    f0 : float
        fundamental frequency
    m : int
        maximum number of the waves, which determines the highest frequency of the components in the synthetic noise

    Returns
    -------
    y : array
        the generated 1/f^alpha noise

    References
    ----------
    Eq. (15) in Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
        Phys Rev E Stat Nonlin Soft Matter Phys 71, 066110 (2005).
    '''
    n = np.size(t)  # number of time points
    y = np.zeros(n)

    if f0 is None:
        f0 = 1/n  # fundamental frequency
    if m is None:
        m = n//2

    k = np.arange(m) + 1  # wave numbers

    rng = np.random.default_rng(seed)
    theta = rng.random(int(m))*2*np.pi  # random phase
    for j in range(n):
        coeff = (k*f0)**(-alpha/2)
        sin_func = np.sin(2*np.pi*k*f0*t[j] + theta)
        y[j] = np.sum(coeff*sin_func)

    return y

def colored_noise_2regimes(alpha1, alpha2, f_break, t, f0=None, m=None, seed=None):
    ''' Generate a colored noise timeseries with two regimes

    Parameters
    ----------
    alpha1, alpha2 : float
        the exponent of the 1/f^alpha noise
    f_break : float
        the frequency where the scaling breaks
    t : float
        time vector of the generated noise
    f0 : float
        fundamental frequency
    m : int
        maximum number of the waves, which determines the highest frequency of the components in the synthetic noise

    Returns
    -------
    y : array
        the generated 1/f^alpha noise

    References
    ----------
     Eq. (15) in Kirchner, J. W. Aliasing in 1/f(alpha) noise spectra: origins, consequences, and remedies.
         Phys Rev E Stat Nonlin Soft Matter Phys 71, 066110 (2005).
    '''
    n = np.size(t)  # number of time points
    y = np.zeros(n)

    if f0 is None:
        f0 = 1/n  # fundamental frequency
    if m is None:
        m = n//2  # so the aliasing is limited

    k = np.arange(m) + 1  # wave numbers

    if seed is not None:
        np.random.seed(seed)

    theta = np.random.rand(int(m))*2*np.pi  # random phase

    f_vec = k*f0
    regime1= k*f0>=f_break
    regime2= k*f0<=f_break
    f_vec1 = f_vec[regime1]
    f_vec2 = f_vec[regime2]
    s = np.exp(alpha1/alpha2*np.log(f_vec1[0])) / f_vec2[-1]

    for j in range(n):
        coeff = np.ndarray((np.size(f_vec)))
        coeff[regime1] = f_vec1**(-alpha1/2)
        coeff[regime2] = (s*f_vec2)**(-alpha2/2)
        sin_func = np.sin(2*np.pi*k*f0*t[j] + theta)
        y[j] = np.sum(coeff*sin_func)

    return y

def arr_str2np(arr):
    arr =  np.array([float(s) for s in arr[1:-1].split(',')])
    return arr

def is_numeric(obj):
    attrs = ['__add__', '__sub__', '__mul__', '__truediv__', '__pow__']
    return all(hasattr(obj, attr) for attr in attrs)

def replace_str(fpath, d):
    ''' Replace the string in a given text file according
    to the dictionary `d`
    '''
    with open(fpath, 'r') as f:
        text = f.read()
        for k, v in d.items():
            search_text = k
            replace_text = v
            text = text.replace(search_text, replace_text)

    with open(fpath, 'w') as f:
        f.write(text)


def download(url: str, fname: str, chunk_size=1024, show_bar=True, verify=False):
    resp = requests.get(url, stream=True, verify=verify)
    total = int(resp.headers.get('content-length', 0))
    if show_bar:
        with open(fname, 'wb') as file, tqdm(
            desc='Fetching data',
            total=total,
            unit='iB',
            unit_scale=True,
            unit_divisor=1024,
        ) as bar:
            for data in resp.iter_content(chunk_size=chunk_size):
                size = file.write(data)
                bar.update(size)
    else:
        with open(fname, 'wb') as file:
            for data in resp.iter_content(chunk_size=chunk_size):
                size = file.write(data)

proxydb_url_dict = {
    'PAGES2kv2': 'https://github.com/fzhu2e/cfr-data/raw/main/pages2kv2.json',
    'pseudoPAGES2k/ppwn_SNRinf_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNRinf_rta.nc',
    'pseudoPAGES2k/ppwn_SNR10_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR10_rta.nc',
    'pseudoPAGES2k/ppwn_SNR2_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR2_rta.nc',
    'pseudoPAGES2k/ppwn_SNR1_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR1_rta.nc',
    'pseudoPAGES2k/ppwn_SNR0.5_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR0.5_rta.nc',
    'pseudoPAGES2k/ppwn_SNR0.25_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR0.25_rta.nc',
    'pseudoPAGES2k/ppwn_SNRinf_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNRinf_fta.nc',
    'pseudoPAGES2k/ppwn_SNR10_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR10_fta.nc',
    'pseudoPAGES2k/ppwn_SNR2_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR2_fta.nc',
    'pseudoPAGES2k/ppwn_SNR1_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR1_fta.nc',
    'pseudoPAGES2k/ppwn_SNR0.5_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR0.5_fta.nc',
    'pseudoPAGES2k/ppwn_SNR0.25_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/ppwn_SNR0.25_fta.nc',
    'pseudoPAGES2k/tpwn_SNR10_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR10_rta.nc',
    'pseudoPAGES2k/tpwn_SNR2_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR2_rta.nc',
    'pseudoPAGES2k/tpwn_SNR1_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR1_rta.nc',
    'pseudoPAGES2k/tpwn_SNR0.5_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR0.5_rta.nc',
    'pseudoPAGES2k/tpwn_SNR0.25_rta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR0.25_rta.nc',
    'pseudoPAGES2k/tpwn_SNR10_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR10_fta.nc',
    'pseudoPAGES2k/tpwn_SNR2_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR2_fta.nc',
    'pseudoPAGES2k/tpwn_SNR1_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR1_fta.nc',
    'pseudoPAGES2k/tpwn_SNR0.5_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR0.5_fta.nc',
    'pseudoPAGES2k/tpwn_SNR0.25_fta': 'https://github.com/fzhu2e/paper-pseudoPAGES2k/raw/main/data/tpwn_SNR0.25_fta.nc',
}

climfd_url_dict = {
    'iCESM_past1000historical/tas': 'https://atmos.washington.edu/~rtardif/LMR/prior/tas_sfc_Amon_iCESM_past1000historical_085001-200512.nc',
    'iCESM_past1000historical/pr': 'https://atmos.washington.edu/~rtardif/LMR/prior/pr_sfc_Amon_iCESM_past1000historical_085001-200512.nc',
    'iCESM_past1000historical/d18O': 'https://atmos.washington.edu/~rtardif/LMR/prior/d18O_sfc_Amon_iCESM_past1000historical_085001-200512.nc',
    'iCESM_past1000historical/psl': 'https://atmos.washington.edu/~rtardif/LMR/prior/psl_sfc_Amon_iCESM_past1000historical_085001-200512.nc',
    'iCESM_past1000/tas': 'https://atmos.washington.edu/~rtardif/LMR/prior/tas_sfc_Amon_iCESM_past1000_085001-184912.nc',
    'iCESM_past1000/pr': 'https://atmos.washington.edu/~rtardif/LMR/prior/pr_sfc_Amon_iCESM_past1000_085001-184912.nc',
    'iCESM_past1000/d18O': 'https://atmos.washington.edu/~rtardif/LMR/prior/d18O_sfc_Amon_iCESM_past1000_085001-184912.nc',
    'iCESM_past1000/psl': 'https://atmos.washington.edu/~rtardif/LMR/prior/psl_sfc_Amon_iCESM_past1000_085001-184912.nc',
    'gistemp1200_ERSSTv4': 'https://github.com/fzhu2e/cfr-data/raw/main/gistemp1200_ERSSTv4.nc.gz',
    'gistemp1200_GHCNv4_ERSSTv5': 'https://data.giss.nasa.gov/pub/gistemp/gistemp1200_GHCNv4_ERSSTv5.nc.gz',
    'CRUTSv4.07/tas': 'https://crudata.uea.ac.uk/cru/data/hrg/cru_ts_4.07/cruts.2304141047.v4.07/tmp/cru_ts4.07.1901.2022.tmp.dat.nc.gz',
    'CRUTSv4.07/pr': 'https://crudata.uea.ac.uk/cru/data/hrg/cru_ts_4.07/cruts.2304141047.v4.07/pre/cru_ts4.07.1901.2022.pre.dat.nc.gz',
    '20CRv3/tas': 'https://downloads.psl.noaa.gov/Datasets/20thC_ReanV3/Monthlies/2mSI-MO/air.2m.mon.mean.nc',
    '20CRv3/pr': 'https://downloads.psl.noaa.gov/Datasets/20thC_ReanV3/Monthlies/sfcSI-MO/prate.mon.mean.nc',
    'HadCRUTv5': 'https://www.metoffice.gov.uk/hadobs/hadcrut5/data/current/analysis/HadCRUT.5.0.1.0.analysis.anomalies.ensemble_mean.nc',
    'HadCRUT4.6_GraphEM': 'https://github.com/fzhu2e/cfr-data/raw/main/HadCRUT4.6_GraphEM_median.nc',
    'GPCCv2020': 'https://downloads.psl.noaa.gov//Datasets/gpcc/monitor/precip.monitor.mon.total.1x1.v2020.nc',
}

ensts_url_dict = {
    'BC09_NINO34': 'https://github.com/fzhu2e/cfr-data/raw/main/BC09_NINO34.csv',
}
